@@ -348,13 +348,13 @@ struct eap_peer_config {
 	 * number of PAC entries to store in a PAC list (default: 10).
 	 *
 	 * fast_pac_format=binary option can be used to select binary format
-	 * for storing PAC entires in order to save some space (the default
+	 * for storing PAC entries in order to save some space (the default
 	 * text format uses about 2.5 times the size of minimal binary format).
 	 *
 	 * crypto_binding option can be used to control PEAPv0 cryptobinding
 	 * behavior:
-	 * 0 = do not use cryptobinding
-	 * 1 = use cryptobinding if server supports it (default)
+	 * 0 = do not use cryptobinding (default)
+	 * 1 = use cryptobinding if server supports it
 	 * 2 = require cryptobinding
 	 */
 	char *phase1;
@@ -408,6 +408,47 @@ struct eap_peer_config {
 	 */
 	char *engine_id;
 
+	/**
+	 * engine2 - Enable OpenSSL engine (e.g., for smartcard) (Phase 2)
+	 *
+	 * This is used if private key operations for EAP-TLS are performed
+	 * using a smartcard.
+	 *
+	 * This field is like engine, but used for phase 2 (inside
+	 * EAP-TTLS/PEAP/FAST tunnel) authentication.
+	 */
+	int engine2;
+
+
+	/**
+	 * pin2 - PIN for USIM, GSM SIM, and smartcards (Phase 2)
+	 *
+	 * This field is used to configure PIN for SIM and smartcards for
+	 * EAP-SIM and EAP-AKA. In addition, this is used with EAP-TLS if a
+	 * smartcard is used for private key operations.
+	 *
+	 * This field is like pin2, but used for phase 2 (inside
+	 * EAP-TTLS/PEAP/FAST tunnel) authentication.
+	 *
+	 * If left out, this will be asked through control interface.
+	 */
+	char *pin2;
+
+	/**
+	 * engine2_id - Engine ID for OpenSSL engine (Phase 2)
+	 *
+	 * "opensc" to select OpenSC engine or "pkcs11" to select PKCS#11
+	 * engine.
+	 *
+	 * This is used if private key operations for EAP-TLS are performed
+	 * using a smartcard.
+	 *
+	 * This field is like engine_id, but used for phase 2 (inside
+	 * EAP-TTLS/PEAP/FAST tunnel) authentication.
+	 */
+	char *engine2_id;
+
+
 	/**
 	 * key_id - Key ID for OpenSSL engine
 	 *
