@@ -1422,8 +1422,10 @@ void eapol_sm_notify_cached(struct eapol_sm *sm)
 {
 	if (sm == NULL)
 		return;
+	wpa_printf(MSG_DEBUG, "EAPOL: PMKSA caching was used - skip EAPOL");
 	sm->SUPP_PAE_state = SUPP_PAE_AUTHENTICATED;
 	sm->suppPortStatus = Authorized;
+	sm->portValid = TRUE;
 	eap_notify_success(sm->eap);
 	eapol_sm_step(sm);
 }
